/*
Lightweight Automated Planning Toolkit
Copyright (C) 2012
Miquel Ramirez <miquel.ramirez@rmit.edu.au>
Nir Lipovetzky <nirlipo@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __ANYTIME_SINGLE_QUEUE_SINGLE_HEURISTIC_GREEDY_BEST_FIRST_SEARCH__
#define __ANYTIME_SINGLE_QUEUE_SINGLE_HEURISTIC_GREEDY_BEST_FIRST_SEARCH__

#include <aptk/at_bfs.hxx>

namespace aptk { namespace search { namespace bfs {

// Anytime Greedy Best-First Search, with one single open list and one single
// heuristic estimator, with delayed evaluation of states generated
template <typename Search_Model, typename Abstract_Heuristic, typename Open_List_Type >
class AT_GBFS_SQ_SH : public AT_BFS_SQ_SH<Search_Model, Abstract_Heuristic, Open_List_Type> {

public:
	typedef typename Search_Model::State_Type State;
	typedef Node< State > Search_Node;
	
	AT_GBFS_SQ_SH(const Search_Model& search_problem)
	:  AT_BFS_SQ_SH<Search_Model, Abstract_Heuristic, Open_List_Type>(search_problem) {}
	
	//! We disregard the node simply if it is already on the close list.
	bool is_closed(Search_Node* n) {  return (this->closed().retrieve(n) != NULL); }

	//! We only take into account h(n)
	virtual void compute_fn(Search_Node *n) { n->fn() = n->hn(); }
};

} } }

#endif
